/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */

#include "qubopt.h"
#include "qub_reportstream.h"
using namespace fq;

//----- Local Functions 
double safelogd( double x );
typedef double (*dfp_fit_eval) (double *params, int n, double x);
typedef void   (*dfp_fit_grads) (double *params, int n, double x, double *grads);

class dfp_fit_result{
	public:
		fq::vector<double> params;
		double sumSquareDiff;
		int iterations;
		int err;

		dfp_fit_result(double* pars, int npar, double ssd, int iter, int errr)
			: params(npar), sumSquareDiff(ssd), iterations(iter), err(errr)
		{
			memcpy(params, pars, npar*sizeof(double));
		}
	};


dfp_fit_result dfp_fit( double* X, double* Y, int ndata,
					    dfp_fit_eval evalF, dfp_fit_grads gradsF, double* initParams, int nparam,
					    int maxIter, double convSSD, double convGrad, double stepMax, int* StopFlag );

// uses dfpmin to minimize sum-square-difference between curve and data
#define IDLFIT_FIT_MAX_ITER       10			// iterations on fitting of the section
#define IDLFIT_FIT_CONVSSD        0.01
#define IDLFIT_FIT_CONVGRAD       0.005
#define IDLFIT_FIT_STEPMAX        0.2			// maximum change as % of grad., consider other values - chris did not yet.  Mil likes 1; MPL like 0.1;

#define FIT_MIN                   4				// min points to try to fit 


class dfp_fitter : public dfp_optimizable
{
public:
	double* X;
	double* Y;
	int N;
	dfp_fit_eval eval;
	dfp_fit_grads grads;
	double* params;
	int nparam;
	int* Stop;

	dfp_fitter(double* XX, double* YY, int ndata,
		       dfp_fit_eval evalF, dfp_fit_grads gradsF, double* initParams, int nparams, int* StopFlag )
		: X(XX), Y(YY), N(ndata),
		  eval( evalF ), grads( gradsF ), params( initParams ), nparam( nparams ), Stop( StopFlag )
	{}

	virtual ~dfp_fitter() {}
  
	virtual bool getStartingZ( fq::vector<double> &z )
	{
		z.resize(nparam);
		memcpy(z, params, nparam*sizeof(double));
		return true;
	}

	virtual bool evaluate( fq::vector<double> &z, fq::vector<double> &gz,
							double &ssd, int &nf, int &ndf, bool flagGrad=true )
	{
		bool rtnVal = false;
		double datY, curveY;

		double* pars = z;
		fq::vector<double> tmpGrads(nparam, 0.0);
		int gradBytes = nparam * sizeof(double);
		memcpy(gz, tmpGrads, gradBytes);

		for ( int i=0; i<N; ++i ) {
			datY = Y[i];
			curveY = eval(pars, nparam, i);
			grads(pars, nparam, i, tmpGrads);
			for ( int p=0; p<nparam; ++p ) {
				gz[p] += 2 * (curveY - datY) * tmpGrads[p];
			}
		}

		if ( rtnVal )
			rtnVal = ! (Stop && *Stop);

		return rtnVal;
	}

	virtual bool checkContinue( int iter, int nf, int ndf, double ssd,
							fq::vector<double> &z, fq::vector<double> &gz )
	{
		return ! (Stop && *Stop);
	}
};


dfp_fit_result dfp_fit( double* X, double* Y, int ndata,
					    dfp_fit_eval evalF, dfp_fit_grads gradsF, double* initParams, int nparam,
					    int maxIter, double convSSD, double convGrad, double stepMax, int* StopFlag )
{
	dfp_fitter fitter(X, Y, ndata, evalF, gradsF, initParams, nparam, StopFlag);
	dfp_result result = dfp_optimize(&fitter, maxIter, convSSD, convGrad, stepMax);
	return dfp_fit_result(result.z, result.z.size(), result.ll, result.iter, result.err);
}


template<int npar>
class IdlFit_Vit_Node
{
public:
	int prev;
	double ll;
	double nextbase;
	int iFit; // data point index in the current fit segment
	double params[npar];
};


template<int npar>
class IdlFit_Vit_Lattice
{
private:
	matrix<char> latticeStorage;

public:
	typedef IdlFit_Vit_Node<npar> Node;
	Node** lattice;
	int nstate, N;
	
	IdlFit_Vit_Lattice(int nstates, int ndata)
		: latticeStorage(nstates, ndata*sizeof(Node)),
		  lattice( (Node**) (char**) latticeStorage ),
		  nstate( nstates ), N( ndata )
	{}
};


double log_gaussian(double mean, double sd, double x)
{
	double tmp = (x - mean) / sd;
 	return log(1.0/(sd*2.5066283)) - .5*tmp*tmp;
}

bool isPowerOfTwo(int x){
	//todo :  return (i|(i-1))+1 == i*2 ;	// 
	while ( x ) {
		if ( x & 1 )
			break;
		x >>= 1;
		}
	x >>= 1;
	if ( x )
		return false;
	else
		return true;
	}


template<int npar, dfp_fit_eval CurveEval, dfp_fit_grads CurveGrads>
class IdlFitter
{
public:
	typedef IdlFit_Vit_Node<npar> Node;
	typedef IdlFit_Vit_Lattice<npar> Lattice;

	IdlFitter( QUB_Tree modelTree, double initbase, double* initPars,
			   int maxIterations, double convergLL, int* StopFlag )
	: model( modelTree ), initBase( initbase ),
	  maxIter( maxIterations ), convLL( convergLL ), Stop( StopFlag ),
	  fitParams(npar)
	{
		memcpy(initParams, initPars, npar*sizeof(double));
	}

	QUB_Tree idealize( double *data, int ndata, int nchannel, double sampMS, int first, int last,
		std::ostream& report, QTR_NodeFptr resultCB )
	{
		if ( model.node.find("States")->find("State")->isNull() || ! ndata )
			return QUB_Tree();

		setupData(data, ndata, nchannel, sampMS);

		Lattice latticeStorage(nstate, ndata);
		lattice = latticeStorage.lattice;
		fq::vector<char> testNodeStorage(nstate * sizeof(Node));
		testNodes = (Node*) (char*) testNodeStorage;
		fq::vector<int> stateSeqStorage(ndata);
		stateSeq = stateSeqStorage;

		double LL, lastLL = 0.0;
		int iter;

		for ( iter=0; iter<maxIter; ++iter ) {
			LL = viterbi();
			reestimate();
			resultCB->NodeCall( buildResult(iter+1, first, last) );

			if ( Stop && *Stop )
				break;
			if ( iter && (LL > lastLL) && ((LL - lastLL) < convLL) )
				break;

			lastLL = LL;
		}

		return buildResult(iter, first, last);
	}


private:
	skm_model model;
	double initBase;
	double initParams[npar];
	int maxIter;
	double convLL;
	int* Stop;

	double *data;
	int ndata;
	int nstate;
	double sampMS;

	fq::vector<double> logPr;
	matrix<double> logA;
	int nclass;
	fq::vector<int> clazz;
	fq::vector<double> amp, sd;

	fq::vector<double> fitX, fitY, fitParams;

	// valid during idealize()
	Node** lattice; // [nmstate, ndata]
	Node* testNodes; // [nmstate]
	int* stateSeq; // [ndata]


	void setupData( double* data, int ndata, int nchannel, double sampMS )
	{
		int i, j;
		double x;

		this->data = data;
		this->ndata = ndata;
		this->sampMS = sampMS;

		skm_metamodel mmodel(model, nchannel, sampMS);
		nstate = mmodel.nmutistate;
		logPr.resize(nstate);
		for ( i=0; i<nstate; ++i )
			logPr[i] = safelogd(mmodel.mutipr[i]);
		logA.resize(nstate, nstate);
		for ( i=0; i<nstate; ++i )
			for ( j=0; j<nstate; ++j )
				logA[i][j] = safelogd(mmodel.mutia[i*nstate+j]);

		nclass = mmodel.nmuticlass;
		clazz = mmodel.cmutistate;
		amp.resize(nclass);
		sd.resize(nclass);
		for ( i=0; i<nstate; ++i ) {
			amp[i] = mmodel.mutiamp[i];
			sd[i] = mmodel.mutixms[i];
		}

		fitX.resize(ndata);
		fitY.resize(ndata);
		for ( i=0, x=0.0; i<ndata; ++i, x+=sampMS )
			fitX[i] = x;
	}

	double viterbi() // fills lattice, stateSeq, fitY=baseline
	{
		int i, j, k, m;
		double x, y;
		Node *node, *prev;

		// i=0
		y = data[0];
		for ( j=0; j<nstate; ++j ) {
			node = lattice[j] + 0;
			node->prev = -1;
			node->ll = logPr[j] + log_gaussian(amp[clazz[j]], sd[clazz[j]], y - initBase);
			node->iFit = 0;
			memcpy(node->params, initParams, npar*sizeof(double));
			// fit(0, j, 0, node->iFit, node->params);
			node->nextbase = CurveEval(node->params, npar, (node->iFit+1) * sampMS);
		}

		// at time i, state j; looking back at time i-1, state k
		for ( i=1; i<ndata; ++i ) {
			y = data[i];
			for ( j=0; j<nstate; ++j ) {
				if ( Stop && *Stop )
					break;

				m = 0; // max ll prev state
				for ( k=0; k<nstate; ++k ) {
					prev = lattice[k] + i - 1;
					node = testNodes + k;

					node->prev = k;
					node->ll = prev->ll + logA[k][j] + log_gaussian(amp[clazz[j]], sd[clazz[j]], y - prev->nextbase);
					node->iFit = (clazz[j] == clazz[k]) ? (prev->iFit + 1) : 0;
					memcpy(node->params, prev->params, npar*sizeof(double));
					if ( (node->iFit >= FIT_MIN) && isPowerOfTwo(node->iFit) )
						fit(i, j, k, node->iFit, node->params);
					node->nextbase = CurveEval(node->params, npar, (node->iFit+1) * sampMS);

					if ( k == 0 )
						x = node->ll;
					else if ( x < node->ll ) {
						node->ll = x;
						m = k;
					}
				}
				
				lattice[j][i] = testNodes[m];
			}
		}
			
		for ( i=1, j=0, x=lattice[0][ndata-1].ll; i<nstate; ++i ) {
			y = lattice[i][ndata-1].ll;
			if ( y > x ) {
				x = y;
				j = i;
			}
		}

		stateSeq[ndata-1] = j;
		fitY[ndata-1] = data[ndata-1] - amp[clazz[j]];
		for ( i=ndata-2; i>=0; --i ) {
			j = lattice[j][i+1].prev;
			stateSeq[i] = j;
			fitY[i] = data[i] - amp[clazz[j]];
		}

		return x;
	}

	void fit( int iDat, int s, int from, int iFit, double* pars )
	{
		int i;
		Node* node;

		fitY[iFit] = data[iDat] - amp[clazz[s]];
		
		if ( iFit ) {
			s = from;
			node = lattice[s] + iDat - 1;
			for ( i=1; i<iFit; ++i ) {
				fitY[iFit - i] = data[iDat-i] - amp[clazz[s]];
				if ( iDat - i ) {
					s = node->prev;
					node = lattice[s] + iDat-i-1;
				}
			}
		}

		double fitpars[npar];
		memcpy(fitpars, pars, npar*sizeof(double));
		dfp_fit_result fitt = dfp_fit( fitX, fitY, iFit+1, CurveEval, CurveGrads, fitpars, 
										npar /* note : changed from 3 to nPar by chris while discussing 6-28-2004 */,
									   IDLFIT_FIT_MAX_ITER, IDLFIT_FIT_CONVSSD, IDLFIT_FIT_CONVGRAD, IDLFIT_FIT_STEPMAX,
									   Stop );
		if ( ! fitt.err ) {
			memcpy(pars, fitpars, npar*sizeof(double));
		}
	}

	void reestimate()
	{
		int i, j, k;
		double x;

		// a matrix:  count transitions, normalize rows, log entries
		for ( j=0; j<nstate; ++j )
			for ( k=0; k<nstate; ++k )
				logA[j][k] = 0.0;

		for ( i=0; i<ndata-1; ++i )
			logA[stateSeq[i]][stateSeq[i+1]] += 1.0;

		for ( j=0; j<nstate; ++j ) {
			x = 0.0;
			for ( k=0; k<nstate; ++k )
				x += logA[j][k];
			if ( x > 0.0 )
				for ( k=0; k<nstate; ++k )
					logA[j][k] /= x;
			for ( k=0; k<nstate; ++k )
				logA[j][k] = safelogd( logA[j][k] );
		}

		// amp, sd
		fq::vector<int> count(nclass, 0);
		fq::vector<double> tmp(nclass, 0.0);
		for ( i=0; i<ndata; ++i ) {
			j = clazz[ stateSeq[i] ];
			tmp[j] += data[i];
			++(count[j]);
		}
		for ( j=0; j<nclass; ++j ) {
			if ( count[j] )
				amp[j] = tmp[j] / count[j];
		}

		for ( j=0; j<nclass; ++j ) {
			tmp[j] = 0.0;
		}
		for ( i=0; i<ndata; ++i ) {
			j = clazz[ stateSeq[i] ];
			x = amp[j] - data[i];
			tmp[j] += x*x;
		}
		for ( j=0; j<nclass; ++j ) {
			if ( count[j] ) {
				x = sqrt( tmp[j] / count[j] );
				if ( x > 1.0e-10 )
					sd[j] = x;
			}
		}
	}

	QUB_Tree buildResult( int iterations, int first, int last )
	{
		QUB_Tree seg = QUB_Tree::Create("Segment");
		seg.setNumData(QTR_TYPE_INT, 2, 1, (void*)0);
		seg.dataAs(0,0,0) = first;
		seg.dataAs(1,0,0) = last;
		seg["amp"].setNumData(QTR_TYPE_DOUBLE, nclass, 1, amp);
		seg["sd"].setNumData(QTR_TYPE_DOUBLE, nclass, 1, sd);

		QUB_Tree fitNode = seg["FitCurve"];
		fitNode.setNumData(QTR_TYPE_DOUBLE, ndata, 1, (void*)0);
		double* fit = (double*) fitNode.data();
		// fill with amp of state seq; then add baseline
		fit[0] = amp[clazz[stateSeq[0]]];

		std::vector<int> idwt, edwt;
		int cls = clazz[stateSeq[0]];
		int i, j;
		for ( i=1; i<ndata; ++i ) {
			j = clazz[stateSeq[i]];
			if ( cls != j ) {
				idwt.push_back(cls);
				edwt.push_back(i-1);
				cls = j;
			}
			fit[i] = amp[clazz[stateSeq[i]]];
		}
		idwt.push_back(cls);
		edwt.push_back(ndata-1);
		int ndwt = idwt.size();

		seg["DwellCount"].setData(QTR_TYPE_INT, ndwt);
		seg["Classes"].setNumData(QTR_TYPE_INT, ndwt, 1, &(idwt[0]));
		QUB_Tree tdwtNode = seg["Durations"];
		tdwtNode.setNumData(QTR_TYPE_FLOAT, ndwt, 1, (void*)0);
		float* tdwt = (float*) tdwtNode.data();
		QUB_Tree sdwtNode = seg["Firsts"];
		sdwtNode.setNumData(QTR_TYPE_INT, ndwt, 1, (void*)0);
		int* sdwt = (int*) sdwtNode.data();
		seg["Lasts"].setNumData(QTR_TYPE_INT, ndwt, 1, &(edwt[0]));

		j = 0;
		for ( i=0; i<ndwt; ++i ) {
			sdwt[i] = j;
			tdwt[i] = (float) (sampMS * (edwt[i] - j + 1));
			j = edwt[i] + 1;
		}

		i = ndata - 1;
		while ( i >= 0 ) {
			Node* node = lattice[stateSeq[i]] + i;
			for ( j=0; j<=node->iFit; ++j )
				fit[i-node->iFit+j] += CurveEval(node->params, 3, j*sampMS);
			i -= node->iFit + 1;
		}

		seg["LL"].setData(QTR_TYPE_DOUBLE, lattice[ stateSeq[ndata-1] ][ ndata-1 ].ll);
		seg["Iterations"].setData(QTR_TYPE_INT, iterations);

		return seg;
	}

};





double ExpEval(double *params, int n, double x)
{
	return ( params[0] * exp( - x / params[1] ) + params[2] );
}

void   ExpGrads(double *params, int n, double x, double *grads)
{
	grads[0] = exp( -x / params[1] );
	grads[1] = grads[0] * params[0] * x / (params[1] * params[1]);
	grads[2] = 1;
}

typedef IdlFitter<3, ExpEval, ExpGrads> IdlFitter_Exp;

/* templates won't compile or tell their errors until you instantiate 'em
void compile4real()
{
	double x;
	int i;

	IdlFitter_Exp fitter(QUB_Tree(), 0.0, &x, 1, 1.0, &i);
	fitter.idealize(&x, 1, 1, 0.05);
}
*/


QUB_Tree IdlFit_Exp( QUB_Tree properties, std::vector<QUB_Tree> datasets,
					 QTR_NodeFptr fillDatCB, QTR_NodeFptr reportCB, QTR_NodeFptr rsltCB )
{
	mil_reportstream report( reportCB );

	fq::vector<double> initPars;
	initPars.push_back( properties["InitA"].dataAsDouble(1.0) );
	initPars.push_back( properties["InitT"].dataAsDouble(1.0) );
	initPars.push_back( properties["InitB"].dataAsDouble(0.0) );

	IdlFitter_Exp fitter(properties["ModelFile"], properties["InitBaseline"].dataAsDouble(), initPars,
		properties["MaxIterations"].dataAsInt(10), properties["ConvLL"].dataAsDouble(0.01),
		(int*) properties["StopFlag"].data());

	QUB_Tree result = QUB_Tree::Create("Result");
	result.appendClone( properties );
	result["ResultType"].setData("IdlFit");
	QUB_TreeIter dsSpot = result.end();

	for ( std::vector<QUB_Tree>::iterator dsi = datasets.begin(); dsi != datasets.end(); ++dsi ) {
		QUB_Tree dataset = *dsi;
		QUB_Tree dsOut = QUB_Tree::Create("DataSet");
		dsSpot.insert(dsOut);
		++dsSpot;
		dsOut.appendClone(dataset["FileName"]);
		dsOut.appendClone(dataset["sampling"]);
		QUB_TreeIter segSpot = dsOut.end();

		int nchannel = (int) dataset["ExpCond"]["ChannelCount"].dataAsDouble(1.0);
		report << "channel count: " << nchannel << endl;
		double sampMS = dataset["sampling"].dataAsDouble(0.1);

		for ( QUB_TreeMonoIter segi = dataset.find("Segment"); ! segi->isNull(); segi.nextSameName() ) {
			int first = segi->dataAs(0, 0, 0);
			int last = segi->dataAs(1, 0, 0);
			int ndata = last - first + 1;
			fillDatCB->NodeCall(*segi);
			double *data = (double*) segi->find("Channel")->data();
			QUB_Tree segResult = fitter.idealize( data, ndata, nchannel, sampMS, segi->dataAs(0,0,0), segi->dataAs(1,0,0), report, rsltCB );
			segi->removeChild( (*segi)["Channel"] );
			segSpot.insert( segResult );
			++segSpot;

			segResult.appendClone( (*segi)["start"] );
			segResult["length"].setData(QTR_TYPE_DOUBLE, ndata*sampMS);
			int ndwt = segResult["DwellCount"].dataAsInt();
			int *firsts = (int*) segResult["Firsts"].data();
			int *lasts = (int*) segResult["Lasts"].data();
			for ( int i=0; i<ndwt; ++i ) {
				firsts[i] += first;
				lasts[i] += first;
			}
		}
	}

	return result;
}

extern "C" MIL_API QTR_Impl * __stdcall
idlfit_tree_iface( QTR_Impl *cfg, QTR_Impl **data, QTR_INodeCFunc *fillDataCB, QTR_INodeCFunc *reportCB, QTR_INodeCFunc *rsltCB ){
	QUB_Tree cfgNode( cfg );

	std::vector<QUB_Tree> dataNodes;
	QTR_Impl **di = data;
	while ( *di )
		dataNodes.push_back( QUB_Tree( *(di++) ) );

	QTR_NodeFptr fillDat, report, rslt;
	if ( fillDataCB )
		fillDat = QTR_NodeFptr( new QTR_NodePasFunc( fillDataCB ) );
	else
		fillDat = QTR_NodeFptr( new QTR_NoNodeFunc );

	if ( reportCB )
		report = QTR_NodeFptr( new QTR_NodePasFunc( reportCB ) );
	else
		report = QTR_NodeFptr( new QTR_NoNodeFunc );

	if ( rsltCB )
		rslt = QTR_NodeFptr( new QTR_NodePasFunc( rsltCB ) );
	else
		rslt = QTR_NodeFptr( new QTR_NoNodeFunc );

	QUB_Tree resultNode = IdlFit_Exp(cfg, dataNodes, fillDat, report, rslt);
	QTR_Impl *result = resultNode.getImpl();
	QTR_INCREF( result );
	return result;
	}

extern "C" PyObject * idlfit_Pyiface( PyObject *self, PyObject *args )
{
	PyObject *cfgImpl, *dataseq, *fillDatPy = 0, *reportPy = 0, *rsltPy = 0;
	if ( ! PyArg_ParseTuple( args, "OO|OOO", &cfgImpl, &dataseq, &fillDatPy, &reportPy, &rsltPy ) )
		return NULL;

	if ( (! qubtree_node_check( cfgImpl )) || (! PySequence_Check( dataseq )) ) {
		PyErr_SetString( PyExc_TypeError, "IdlFit(cfgTree, [idlTree1, idlTree2, ...])" );
		return NULL;
	}

	QUB_Tree cfg = qubtree_node_obj( cfgImpl );

	std::vector<QUB_Tree> data( PySequence_Size( dataseq ) );
	if ( ! data.size() ) {
		PyErr_SetString( PyExc_TypeError, "IdlFit needs at least one data set" );
		return NULL;
	}
	for ( unsigned int i=0; i<data.size(); ++i ) {
		PyObject *dat = PySequence_GetItem( dataseq, i );
		if ( qubtree_node_check( dat ) ) {
			data[i] = qubtree_node_obj( dat );
			Py_DECREF( dat );
		}
		else {
			Py_DECREF( dat );
			PyErr_SetString( PyExc_TypeError, "arg 2 must be a sequence of DataSet trees" );
			return NULL;
		}
	}

	QTR_NodeFptr fillDataCB, reportCB, rsltCB;
	if ( fillDatPy )
		fillDataCB = QTR_NodeFptr( new QTR_NodePyFunc( fillDatPy ) );
	else
		fillDataCB = QTR_NodeFptr( new QTR_NoNodeFunc );

	if ( reportPy )
		reportCB = QTR_NodeFptr( new QTR_NodePyFunc( reportPy ) );
	else
		reportCB = QTR_NodeFptr( new QTR_NoNodeFunc );

	if ( rsltPy )
		rsltCB = QTR_NodeFptr( new QTR_NodePyFunc( rsltPy ) );
	else
		rsltCB = QTR_NodeFptr( new QTR_NoNodeFunc );

	QUB_Tree result = IdlFit_Exp(cfg, data, fillDataCB, reportCB, rsltCB);

	return qubtree_node_using( result.getImpl() );
}

//-----------------------------------------------------------------------
double safelogd( double x ){
	return ( (x > 0.0) ? log(x) : -1.0e10 );
	}
